var redis = require("redis");

var CacheMiddleware = function (options) {
	if (options.client){
		this.client = options.client;
	} else {
		this.client = options.redisConnString ? redis.createClient(options.redisConnString) : redis.createClient();
	}
	this.cacheDuration = options.cacheDuration || 60*5;
	this.cacheControl = options.cacheControl || "public";
	this.cacheExceptions = options.cacheExceptions || [];
}

CacheMiddleware.prototype.set = function(req, res, html, callback) {
	var self = this;
	var key = req.path;
	
	self.client.set(key, html, function(err, result) {
		self.client.expire(key, self.cacheDuration, function(e, r) {			
			self.setHeaders(req, res, callback);
		});
	});	
};

CacheMiddleware.prototype.shouldCache = function(req) {
	var self = this;

	if (self.cacheExceptions.length > 0) {
		for (var i = 0, l = self.cacheExceptions.length; i < l; i++) {
			var rg = new RegExp(self.cacheExceptions[i]);
			
			if (rg.test(req.path)) {
				return false;
			}
		}
	}

	return true;
};

CacheMiddleware.prototype.setHeaders = function(req, res, callback) {
	var self = this;
	res.setHeader("Cache-Control", self.cacheControl + ", max-age="+self.cacheDuration);
	var d = new Date();
	self.client.ttl(req.path, function(err, val){
		d.setSeconds(d.getSeconds()+val);
		res.setHeader("Expires", d);
		callback()	
	});
};

CacheMiddleware.prototype.get = function(req, res, callback) {
	var self = this;
	var key = req.path;
	self.client.get(key, function(err, result) {
		self.setHeaders(req, res, function(){
			callback(err, result);	
		});
	});
};

var _CacheMiddleware;

var cache = function(req, res, next) {
	if (!_CacheMiddleware.shouldCache(req)) {
		next();
	} else {
	    var oldWrite = res.write,
	        oldEnd = res.end;
	
		_CacheMiddleware.get(req, res, function(err, html){
			if (html) {
		      	res.send(html);
			} else {		
	    	  var chunks = [];

	    	  res.write = function (chunk) {
	    	    chunks.push(chunk);

	    	    oldWrite.apply(res, arguments);
	    	  };

	    	  res.end = function (chunk, encoding) {
	    	    if (chunk)
	    	      chunks.push(chunk);

	    	    var body = Buffer.concat(chunks).toString('utf8');
				_CacheMiddleware.set(req, res, body, function() {
					res.end = oldEnd;
					res.end(chunk, encoding);
				});
	    	  };

	    	  next();			
			}	
		});
	}
};

module.exports = {
	createCache: function(options) {
		_CacheMiddleware = new CacheMiddleware(options || {});
		
		return cache;
	}
}